package be.iminds.ilabt.jfed.espec.filefetcher;

import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * ExperimentSpecificationFileManager
 *   Fetches all files in an ExperimentSpecification, putting them in memory.
 *   It downloads any files needing download in parallell,
 *   and has a method to wait for all files to be ready.
 */
public class ExperimentSpecificationFileManager {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationFileManager.class);

    @Nonnull private final ExperimentSpecification eSpec;
    @Nonnull private final ESpecBundle bundle;
    @Nonnull private final ESpecLogger eSpecLogger;

    private final Map<FileSource, byte[]> bytes = new ConcurrentHashMap<>();
    private final ExecutorService executorService;

    private boolean started;
    private boolean errorOccured;

    private AtomicInteger slowFetchCountdown;

    public ExperimentSpecificationFileManager(@Nonnull ExperimentSpecification eSpec, @Nonnull ESpecBundle bundle, @Nonnull ESpecLogger eSpecLogger) {
        this.eSpec = eSpec;
        this.bundle = bundle;
        this.eSpecLogger = eSpecLogger;
        this.executorService = Executors.newFixedThreadPool(3);
        this.errorOccured = false;
    }

    public void fetchAll() {
        eSpecLogger.firePreFileLoad();
        started = true;

        List<FileSource> fileSources = new ArrayList<>();
        eSpec.getUploads().stream().map(UploadSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
        eSpec.getExecutes().stream().map(ExecuteSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
        if (eSpec.getAnsible() != null) {
            eSpec.getAnsible().getAnsibleHostSpec().getUploadLists().stream()
                    .map(UploadSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
            eSpec.getAnsible().getAnsibleHostSpec().getExecuteLists().stream()
                    .map(ExecuteSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
            eSpec.getAnsible().getAnsibleGalaxySpecs().stream()
                    .map(UploadLikeSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
            eSpec.getAnsible().getAnsiblePlaybookSpecs().stream()
                    .map(UploadLikeSpec::getSource).filter(Objects::nonNull).forEach(s -> fileSources.add(s));
        }

        Map<Boolean, List<FileFetcher>> fileSourceListBySpeed =
                fileSources.stream()
                        .map(fileSource -> new FileFetcher(fileSource, bundle))
                        .collect(Collectors.partitioningBy(FileFetcher::isFast));

        List<FileFetcher> fastFetchers = fileSourceListBySpeed.get(true);
        List<FileFetcher> slowFetchers = fileSourceListBySpeed.get(false);

        for (FileFetcher fff : fastFetchers) {
            //generated keypair is handled in a different way
            if (fff.getFileSource().isGeneratedKeyPair())
                continue;

            try {
                byte[] b = fff.fetchBytes();
                assert b != null;
                LOG.debug("ExperimentSpecificationFileManager fetched "+b.length+" bytes for "+fff.getFileSource().getBasename());
                bytes.put(fff.getFileSource(), b);
                eSpecLogger.firePostFileLoad(fff.getFileSource(), true, b.length, true);
            } catch (IOException e) {
                LOG.error("Failed to (fast) fetch file with source type "+fff.getFileSource().getType()+"", e);
                errorOccured = true;
                executorService.shutdown();
                eSpecLogger.firePostFileLoad(fff.getFileSource(), true, 0, false);
                eSpecLogger.firePostFileLoadAll(false);
                return;
            }
        }

        LOG.debug("ExperimentSpecificationFileManager fetched "+fastFetchers.size()+" fast files.");

        if (!slowFetchers.isEmpty()) {
            slowFetchCountdown = new AtomicInteger(slowFetchers.size());
            for (FileFetcher sff : slowFetchers) {
                //generated keypair is handled differently
                if (sff.getFileSource().isGeneratedKeyPair())
                    continue;

                queueFileFetch(sff);
            }

            LOG.debug("ExperimentSpecificationFileManager queued " + slowFetchers.size() + " slow files for fetching.");
        } else {
            LOG.debug("ExperimentSpecificationFileManager has no slow files to fetch.");
            eSpecLogger.firePostFileLoadAll(true);
        }

        //always shutdown, even if nothing was added.
        //   (otherwise, awaitTermination and isTerminated will not work as wanted)
        executorService.shutdown();
    }

    private void queueFileFetch(@Nonnull FileFetcher fileFetcher) {
        Runnable backGroundFetch = new Runnable() {
            @Override
            public void run() {
                try {
                    byte[] b = fileFetcher.fetchBytes();
                    eSpecLogger.firePostFileLoad(fileFetcher.getFileSource(), false, b.length, true);
                    bytes.put(fileFetcher.getFileSource(), b);
                } catch (IOException e) {
                    LOG.error("Failed to (slow) fetch file with source type "+
                            fileFetcher.getFileSource().getType()+"", e);
                    errorOccured = true;
                    eSpecLogger.firePostFileLoad(fileFetcher.getFileSource(), true, 0, false);
                    executorService.shutdownNow();
                    eSpecLogger.firePostFileLoadAll(false);
                }

                if (slowFetchCountdown.decrementAndGet() == 0 && !errorOccured) {
                    eSpecLogger.firePostFileLoadAll(true);
                }
            }
        };
        executorService.execute(backGroundFetch);
    }

    public boolean isReady() {
        if (!started)
            throw new IllegalStateException("Dangerous to call isReady before starting!");
        return executorService.isTerminated();
    }

    public boolean getErrorOccured() {
        return errorOccured;
    }

    /**
     * Wait for up to 30 minutes for all tasks to become ready
     *
     * @throws InterruptedException
     */
    public void waitUntilReady() throws InterruptedException {
        if (!started)
            throw new IllegalStateException("Dangerous to call waitUntilReady before starting!");

        if (isReady()) {
            LOG.debug("ExperimentSpecificationFileManager did not have to wait for all files to be ready");
            return;
        }

        LOG.debug("ExperimentSpecificationFileManager waits for all files to be ready");

        boolean terminated = this.executorService.awaitTermination(30, TimeUnit.MINUTES);
        if (!terminated)
            throw new IllegalStateException("Failed to wait until all files are ready");

        LOG.debug("ExperimentSpecificationFileManager stopped waiting for files to be ready ready");
    }

    @Nonnull
    public byte[] getBytes(@Nonnull FileSource fileSource) {
        if (!isReady())
            throw new IllegalStateException("ExperimentSpecificationFileManager requires you to wait until all files are ready, before using them.");

        byte[] res =  bytes.get(fileSource);
        if (res == null) {
            throw new IllegalStateException("Unexpectedly did not have content of " + fileSource.getBasename() + " (debug details: search fileSource="+fileSource+". Known fileSources="+bytes.keySet()+")");
        }
        return res;
    }



    @Nonnull
    public ExperimentSpecification getExperimentSpecification() {
        return eSpec;
    }

    @Nonnull
    public ESpecBundle getBundle() {
        return bundle;
    }
}
